// Copyright 1998-2017 Epic Games, Inc. All Rights Reserved.

/*=============================================================================
	MetalRHIPrivate.h: Private Metal RHI definitions.
=============================================================================*/

#pragma once

#include "CoreMinimal.h"
#include "Misc/ScopeLock.h"
#include "Misc/CommandLine.h"

// UE4 has a Max of 8 RTs, but we can spend less time looping with 6
const uint32 MaxMetalRenderTargets = 6;

// Requirement for vertex buffer offset field
const uint32 BufferOffsetAlignment = 256;

// The buffer page size that can be uploaded in a set*Bytes call
const uint32 MetalBufferPageSize = 4096;

#define METAL_API_1_1 (__IPHONE_9_0 || __MAC_10_11)
#define METAL_API_1_2 ((__IPHONE_10_0 && defined(__IPHONE_OS_VERSION_MIN_REQUIRED) && __IPHONE_OS_VERSION_MIN_REQUIRED >= __IPHONE_10_0) || (__MAC_10_12 && defined(__MAC_OS_X_VERSION_MIN_REQUIRED) && __MAC_OS_X_VERSION_MIN_REQUIRED >= __MAC_10_12))

#if METAL_API_1_1
#define BUFFER_CACHE_MODE MTLResourceCPUCacheModeWriteCombined
#else
#define BUFFER_CACHE_MODE MTLResourceOptionCPUCacheModeWriteCombined
#endif

#if PLATFORM_MAC
#define BUFFER_MANAGED_MEM MTLResourceStorageModeManaged
#define BUFFER_STORAGE_MODE MTLStorageModeManaged
#define BUFFER_RESOURCE_STORAGE_MANAGED MTLResourceStorageModeManaged
#define BUFFER_DYNAMIC_REALLOC BUF_AnyDynamic
// How many possible vertex streams are allowed
const uint32 MaxMetalStreams = 31;
#else
#define BUFFER_MANAGED_MEM 0
#define BUFFER_STORAGE_MODE MTLStorageModeShared
#define BUFFER_RESOURCE_STORAGE_MANAGED MTLResourceStorageModeShared
#define BUFFER_DYNAMIC_REALLOC BUF_Volatile
// How many possible vertex streams are allowed
const uint32 MaxMetalStreams = 30;
#endif

#ifndef METAL_STATISTICS
#define METAL_STATISTICS 0
#endif

#define METAL_DEBUG_OPTIONS !(UE_BUILD_SHIPPING || UE_BUILD_TEST)
#if METAL_DEBUG_OPTIONS
#define METAL_DEBUG_OPTION(Code) Code
#else
#define METAL_DEBUG_OPTION(Code)
#endif

#define SHOULD_TRACK_OBJECTS (UE_BUILD_DEBUG)

#define UNREAL_TO_METAL_BUFFER_INDEX(Index) ((MaxMetalStreams - 1) - Index)
#define METAL_TO_UNREAL_BUFFER_INDEX(Index) ((MaxMetalStreams - 1) - Index)

// Dependencies
#include "MetalRHI.h"
#include "RHI.h"
#import <Metal/Metal.h>
#import <QuartzCore/CAMetalLayer.h>

#if !METAL_API_1_1
#define MTLVisibilityResultModeCounting ((MTLVisibilityResultMode)2)
#endif

#if !METAL_API_1_2
#define MTLFeatureSet_iOS_GPUFamily3_v1 ((MTLFeatureSet)4)
#define MTLFeatureSet_iOS_GPUFamily1_v3 ((MTLFeatureSet)5)
#define MTLFeatureSet_iOS_GPUFamily2_v3 ((MTLFeatureSet)6)
#define MTLFeatureSet_iOS_GPUFamily3_v2 ((MTLFeatureSet)7)
#define MTLFeatureSet_tvOS_GPUFamily1_v2 ((MTLFeatureSet)30001)
#define MTLFeatureSet_OSX_GPUFamily1_v2 ((MTLFeatureSet)10001)
#define MTLPixelFormatDepth16Unorm ((MTLPixelFormat)250)
#define MTLPixelFormatX24_Stencil8 ((MTLPixelFormat)262)
#define MTLPixelFormatX32_Stencil8 ((MTLPixelFormat)261)
#endif

// Access the internal context for the device-owning DynamicRHI object
FMetalDeviceContext& GetMetalDeviceContext();

// Safely release a metal resource, correctly handling the case where the RHI has been destructed first
void SafeReleaseMetalResource(id Object);

// Safely release a pooled buffer, correctly handling the case where the RHI has been destructed first
void SafeReleasePooledBuffer(id<MTLBuffer> Buffer);

// Access the underlying surface object from any kind of texture
FMetalSurface* GetMetalSurfaceFromRHITexture(FRHITexture* Texture);

#define NOT_SUPPORTED(Func) UE_LOG(LogMetal, Fatal, TEXT("'%s' is not supported"), L##Func);

#if SHOULD_TRACK_OBJECTS
void TrackMetalObject(NSObject* Object);
void UntrackMetalObject(NSObject* Object);
#define TRACK_OBJECT(Stat, Obj) INC_DWORD_STAT(Stat); TrackMetalObject(Obj)
#define UNTRACK_OBJECT(Stat, Obj) DEC_DWORD_STAT(Stat); UntrackMetalObject(Obj)
#else
#define TRACK_OBJECT(Stat, Obj) INC_DWORD_STAT(Stat)
#define UNTRACK_OBJECT(Stat, Obj) DEC_DWORD_STAT(Stat)
#endif

enum EMetalIndexType
{
	EMetalIndexType_None   = 0,
	EMetalIndexType_UInt16 = 1,
	EMetalIndexType_UInt32 = 2
};

FORCEINLINE MTLIndexType GetMetalIndexType(EMetalIndexType IndexType)
{
	switch (IndexType)
	{
		case EMetalIndexType_UInt16: return MTLIndexTypeUInt16;
		case EMetalIndexType_UInt32: return MTLIndexTypeUInt32;
		case EMetalIndexType_None:
		{
			UE_LOG(LogMetal, Fatal, TEXT("There is not equivalent MTLIndexType for EMetalIndexType_None"));
			return MTLIndexTypeUInt16;
		}
	}
}

FORCEINLINE EMetalIndexType GetRHIMetalIndexType(MTLIndexType IndexType)
{
	switch (IndexType)
	{
		case MTLIndexTypeUInt16: return EMetalIndexType_UInt16;
		case MTLIndexTypeUInt32: return EMetalIndexType_UInt32;
	}
}

FORCEINLINE int32 GetMetalCubeFace(ECubeFace Face)
{
	// According to Metal docs these should match now: https://developer.apple.com/library/prerelease/ios/documentation/Metal/Reference/MTLTexture_Ref/index.html#//apple_ref/c/tdef/MTLTextureType
	switch (Face)
	{
		case CubeFace_PosX:;
		default:			return 0;
		case CubeFace_NegX:	return 1;
		case CubeFace_PosY:	return 2;
		case CubeFace_NegY:	return 3;
		case CubeFace_PosZ:	return 4;
		case CubeFace_NegZ:	return 5;
	}
}

FORCEINLINE MTLLoadAction GetMetalRTLoadAction(ERenderTargetLoadAction LoadAction)
{
	switch(LoadAction)
	{
		case ERenderTargetLoadAction::ENoAction: return MTLLoadActionDontCare;
		case ERenderTargetLoadAction::ELoad: return MTLLoadActionLoad;
		case ERenderTargetLoadAction::EClear: return MTLLoadActionClear;
		default: return MTLLoadActionDontCare;
	}
}

FORCEINLINE MTLStoreAction GetMetalRTStoreAction(ERenderTargetStoreAction StoreAction)
{
	switch(StoreAction)
	{
		case ERenderTargetStoreAction::ENoAction: return MTLStoreActionDontCare;
		case ERenderTargetStoreAction::EStore: return MTLStoreActionStore;
		case ERenderTargetStoreAction::EMultisampleResolve: return MTLStoreActionMultisampleResolve;
		default: return MTLStoreActionDontCare;
	}
}

uint32 TranslateElementTypeToSize(EVertexElementType Type);

MTLPrimitiveType TranslatePrimitiveType(uint32 PrimitiveType);

#include "MetalStateCache.h"
#include "MetalContext.h"

