// Copyright 1998-2017 Epic Games, Inc. All Rights Reserved.

#include "Delay.h"
#include "Dsp.h"
#include "AudioMixer.h"

namespace Audio
{
	FDelay::FDelay()
		: AudioBuffer(nullptr)
		, AudioBufferSize(0)
		, ReadIndex(0)
		, WriteIndex(0)
		, SampleRate(0)
		, DelayInSamples(0.0f)
		, DelayMsec(0.0f)
		, OutputAttenuation(1.0f)
		, OutputAttenuationDB(0.0f)
	{
		Reset();
	}

	FDelay::~FDelay()
	{
		if (AudioBuffer)
		{
			delete[] AudioBuffer;
		}
	}

	void FDelay::Init(const int32 InSampleRate, const int32 InBufferSizeSamples)
	{
		SampleRate = InSampleRate;
		AudioBufferSize = InBufferSizeSamples;

		if (AudioBuffer)
		{
			delete[] AudioBuffer;
		}

		AudioBuffer = new float[InBufferSizeSamples];

		Reset();
	}

	void FDelay::Reset()
	{
		if (AudioBuffer)
		{
			FMemory::Memzero(AudioBuffer, sizeof(float)*AudioBufferSize);
		}

		WriteIndex = 0;
		ReadIndex = 0;

		Update();
	}

	void FDelay::SetDelayMsec(const float InDelayMsec)
	{
		DelayMsec = InDelayMsec;
		Update();
	}

	void FDelay::SetOutputAttenuationDB(const float InDelayAttenDB)
	{
		OutputAttenuationDB = InDelayAttenDB;
		Update();
	}

	float FDelay::Read() const
	{
		// Read the output of the delay at ReadIndex
		const float Yn = AudioBuffer[ReadIndex];

		// Read the location ONE BEHIND yn at y(n-1)
		int32 ReadIndexPrev = ReadIndex - 1;
		if (ReadIndexPrev < 0)
		{
			ReadIndexPrev = AudioBufferSize - 1;
		}

		// Set y(n-1)
		const float YnPrev = AudioBuffer[ReadIndexPrev];

		// Get the amount of fractional delay between previous and next read indices
		const float Fraction = DelayInSamples - (int32)DelayInSamples;

		return FMath::Lerp(Yn, YnPrev, Fraction);
	}

	float FDelay::ReadDelayAt(const float InReadMsec) const
	{
		const float ReadAtDelayInSamples = InReadMsec*((float)SampleRate) / 1000.0f;

		// Subtract to make read index
		int32 ReadAtReadIndex = WriteIndex - (int32)ReadAtDelayInSamples;

		if (ReadAtReadIndex < 0)
		{
			ReadAtReadIndex += AudioBufferSize;	// amount of wrap is Read + Length
		}

		// Read the output of the delay at ReadAtReadIndexs
		float Yn = AudioBuffer[ReadAtReadIndex];

		// Read the location ONE BEHIND yn at y(n-1)
		int32 ReadAtReadIndexPrev = ReadAtReadIndex - 1;
		if (ReadAtReadIndexPrev < 0)
		{
			ReadAtReadIndexPrev = AudioBufferSize - 1;
		}

		// get y(n-1)
		const float YnPrev = AudioBuffer[ReadAtReadIndexPrev];

		// interpolate: (0, yn) and (1, yn_1) by the amount fracDelay
		float Fraction = ReadAtDelayInSamples - (int32)ReadAtDelayInSamples;

		return FMath::Lerp(Yn, YnPrev, Fraction);
	}

	void FDelay::WriteDelayAndInc(const float InDelayInput)
	{
		// write to the delay line
		AudioBuffer[WriteIndex] = InDelayInput; // external feedback sample
												// increment the pointers and wrap if necessary
		WriteIndex++;
		if (WriteIndex >= AudioBufferSize)
		{
			WriteIndex = 0;
		}

		ReadIndex++;
		if (ReadIndex >= AudioBufferSize)
		{
			ReadIndex = 0;
		}
	}

	void FDelay::ProcessAudio(const float* InputSample, float* OutputSample)
	{
		const float Xn = *InputSample;
		const float Yn = DelayInSamples == 0 ? Xn : Read();
		WriteDelayAndInc(Xn);
		*OutputSample = OutputAttenuation*Yn;
	}

	void FDelay::Update()
	{
		// Compute linear output attenuation based on DB attenuation settings
		OutputAttenuation = FMath::Pow(10.0f, OutputAttenuationDB / 20.0f);

		// Compute the delay in samples based on msec delay line
		DelayInSamples = DelayMsec*((float)SampleRate / 1000.0f);

		if (DelayInSamples >= (float)AudioBufferSize)
		{
			DelayInSamples = (float)AudioBufferSize;
		}

		// Subtract from write index the delay in samples (will do interpolation during read)
		ReadIndex = WriteIndex - (int32)DelayInSamples;

		// If negative, wrap around
		if (ReadIndex < 0)
		{
			ReadIndex += AudioBufferSize;
		}
	}

}
